#pragma once

#include <list>
#include <mutex>
#include <thread>
#include <condition_variable>
#include <iostream>
#include <memory>
#include <atomic>
#include <functional>

#include"util.h"

namespace gochat {

	template <typename T>
	class SyncQueue{
	public:
		SyncQueue(int maxSize) : max_size_(maxSize), need_stop(false) { }
		//¼
		void Put(const T& x);
		//¼
		void Put(T && x);
		//Ӷȡ¼,ȡ¼
		void Take(std::list<T> &list);
		//ȡһ¼
		void Take(T &t);
		//ֹͬ
		void Stop();
		//Ϊ
		bool Empty();
		//Ϊ
		bool Full();
		//дС
		size_t Size();
		//дС
		int Count();
	private:
		//вΪ
		bool NotFull() const;
		//вΪ
		bool NotEmpty() const;

		void Add(T && x);
		//
		std::list<T> queue_;
		//
		std::mutex mutex_;
		//вΪյ
		std::condition_variable not_empty_;
		//вΪ
		std::condition_variable not_full_;
		//󳤶
		int max_size_;
		//ֹıʶ,ΪtrueʱͬҪֹ
		bool need_stop;

	};


	class ThreadPool{
	public:
		using Task = std::function<void()>;

		ThreadPool(int numThreads = std::thread::hardware_concurrency()) :queue_(kMaxTaskCount) {
			Start(numThreads);
		}

		~ThreadPool();

		//ֻ֤̻߳һStopThreadGroup
		void Stop();

		//,ֵת
		void AddTask(Task && task);

	private:
		//numThreads߳
		void Start(int numThreads);
		//ȡеȫ,ִ
		void RunInThread();
		//ִֹ
		void StopThreadGroup();
		//߳
		std::list<std::shared_ptr<std::thread>> thread_group;
		//ͬ
		SyncQueue<Task> queue_;
		//еı־,flaseֹ
		std::atomic_bool running_;
		//֤ںڶֻ̻߳һ
		std::once_flag flag_;
	};

}

